/**
 * Copyright (C) 2009-2010 the original author or authors.
 * See the notice.md file distributed with this work for additional
 * information regarding copyright ownership.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.fusesource.scalate
package maven

import collection.mutable.ListBuffer

import java.io.File
import java.net.URLClassLoader
import java.util.ArrayList;

import org.apache.maven.artifact.Artifact
import org.apache.maven.plugin.AbstractMojo
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;

import org.fusesource.scalate.servlet.ServletRenderContext;
import org.fusesource.scalate.util.{ClassPathBuilder, IOUtil, FileResourceLoader}

import org.scala_tools.maven.mojo.annotations._


/**
 * This goal builds precompiles the Scalate templates
 * as Scala source files that be included in your standard
 * build. 
 *
 * @author <a href="http://hiramchirino.com">Hiram Chirino</a>
 */
@goal("precompile")
@phase("process-classes")
@requiresProject
@requiresDependencyResolution("compile")
class PrecompileMojo extends AbstractMojo {

  @parameter
  @expression("${project}")
  @readOnly
  @required
  var project: MavenProject = _

  @parameter
  @description("The directory where the templates files are located.")
  @expression("${basedir}/src/main/webapp")
  var warSourceDirectory: File = _

  @parameter
  @description("The directory where resources are located.")
  @expression("${basedir}/src/main/resources")
  var resourcesSourceDirectory: File = _

  @parameter
  @description("The directory where the scala code will be generated into.")
  @expression("${project.build.directory}/generated-sources/scalate")
  var targetDirectory: File = _

  @parameter
  @description("The directory containing generated classes .")
  @expression("${project.build.outputDirectory}")
  var classesDirectory:File = _

  @parameter
  @description("Additional template paths to compile.")
  var templates:ArrayList[String] = new ArrayList[String]()

  def execute() = {
    targetDirectory.mkdirs();

    getLog.debug("targetDirectory: " + targetDirectory)
    getLog.debug("warSourceDirectory: " + warSourceDirectory)
    getLog.debug("resourcesSourceDirectory: " + resourcesSourceDirectory)

    var engine = new TemplateEngine() {
      // lets output generated bytecode to the classes directory.
      override def bytecodeDirectory = classesDirectory
    }
    
    engine.bindings = createBindings()
    engine.resourceLoader = new FileResourceLoader(Some(warSourceDirectory));
    engine.classpath = getProjectClassPath

    getLog.debug("scalac classpath set to: "+engine.classpath)
    
    engine.workingDirectory = targetDirectory

    val sourceDirs = List(warSourceDirectory, resourcesSourceDirectory)
    var paths = List[String]()
    for (extension <- engine.codeGenerators.keysIterator; sd <- sourceDirs if sd.exists) {
      paths = collectUrisWithExtension(sd, "", "." + extension) ::: paths;
    }

    import collection.JavaConversions._
    templates.foreach { x=>
      paths ::= x 
    }

    getLog.info("Precompiling Scalate Templates into Scala classes...");

    val loader = createProjectClassLoader
    Thread.currentThread.setContextClassLoader(loader)

    for (uri <- paths) {

      // TODO it would be easier to just generate Source + URI pairs maybe rather than searching again for the source file???
      val file = sourceDirs.map(new File(_, uri)).find(_.exists).getOrElse(uri)
      getLog.info("    processing " + file)

      val template = engine.load(uri)
    }
  }

  protected def collectUrisWithExtension(basedir: File, baseuri: String, extension: String): List[String] = {
    var collected = List[String]()
    if (basedir.isDirectory()) {
      var files = basedir.listFiles();
      if (files != null) {
        for (file <- files) {
          if (file.isDirectory()) {
            collected = collectUrisWithExtension(file, baseuri + "/" + file.getName(), extension) ::: collected;
          } else {
            if (file.getName().endsWith(extension)) {
              collected = baseuri + "/" + file.getName() :: collected;
            } else {
            }

          }
        }
      }
    }
    collected
  }


  protected def createBindings():List[Binding] = {
    List(Binding("context", classOf[ServletRenderContext].getName, true, isImplicit = true))
  }

  protected  def createBindingsForPath(uri:String): List[Binding] = {
    Nil
  }

  def getProjectClassPathList() = {
    val classpath = ListBuffer[String]()
    classpath += classesDirectory.getAbsolutePath()

    val classpathArtifacts = project.getArtifacts().asInstanceOf[java.util.Set[Artifact]]
    import collection.JavaConversions._
    classpathArtifacts.foreach { artifact =>
      if (artifact.getArtifactHandler().isAddedToClasspath()) {
        val file = artifact.getFile();
        if (file != null) {
          classpath.add(file.getPath());
        }
      }
    }
    classpath.toList
  }

  def getProjectClassPath() = {
    val builder = new ClassPathBuilder
    getProjectClassPathList.foreach { x=>
      builder.addEntry(x)
    }
    builder.classPath
  }

  def createProjectClassLoader() = {
    val urls = getProjectClassPathList().map { x=>
      new File(x).toURL
    }.toArray

    new URLClassLoader(urls, null)
  }

}
